// command.c

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "command.h"
#include "serial.h"
#include "delay_timer.h"
#include "xmodem.h"
#include "..\flash\flash_header.h"

// external reference to the
// structure contained in 'flash_converted.c' which contains
// the code which is loaded into RAM and executed from RAM
extern const struct rom_data flashcode;

void InitCommandHandler (void)
{
	InitSci();
	InitDelayTimer();
	
	// copy all RAM based code to RAM
	memcpy( (void *) flashcode.start_address, &flashcode.data[0], flashcode.data_length );
}

//
// functionality of this handler will be dependent on the communication hardware interface
//
void RunCommandHandler (void)
{
	unsigned long *ptr;
	pt2FunctionErase fp;		// function pointer used to call target program
	union union_c2s c2s;
	
	// check if bootloader should run by detecting activity for a 3 seconds on the
	// UART channel
	c2s.us = GetByte (3000);
	
	if ( c2s.uc[0] == TIMEOUT )
	{
		// timed out - call target app if reset vector is valid
		ptr = (unsigned long *)( FIRST_USER_FLASH_ADDR + RESET_VECTOR_OFFSET );
		if ( *ptr != 0xffffffff )
		{
			// call user program and never return
			fp = (pt2FunctionErase) *ptr;
			fp( 0, 0);
		}
		else
		{
			// return back to main
			return;
		}
	}
	else
	{
		// serial activity detected
		// clear buffer
		// loop until we get a timeout
		PurgeComms( 250 );	
		// display menu
		ShowMenu();
			
		while (1)
		{
			// timeout is not important as we know we have a byte waiting for us
			c2s.us = GetByte (100000);
			if ( c2s.uc[0] == OK )
			{
				switch ( c2s.uc[1] )
				{
					case '1':
						Command_1();		// blank check
					break;
					
					case '2':
						Command_2();		// erase all user flash blocks
					break;
					
					case '3':
						Command_3();		// erase specific flash block
					break;
					
					case '4':
						Command_4();		// xmodem download to flash memory
					break;

					case '5':
						Command_5();		// run target application
					break;
					
					default:
						ShowMenu();
				}
			}
		}
	}
}

void ShowMenu (void)
{
	SendLFCR();
	SendLFCR();
	SendString( "M16C/62P Bootloader Main Menu v1.0" );
	SendLFCR();
	SendString( "----------------------------------"  );
	SendLFCR();
	SendLFCR();
	SendString( "1......Blank Check User Area" );
	SendLFCR();
	SendLFCR();
	SendString( "2......Erase User Area" );
	SendLFCR();
	SendLFCR();
	SendString( "3......Erase User Block" );
	SendLFCR();
	SendLFCR();
	SendString( "4......Program Flash via XModem Download" );
	SendLFCR();
	SendLFCR();
	SendString( "5......Start User Program" );
	SendLFCR();
	SendLFCR();
	SendByte( '>' );
}

void Command_1 (void)
{
	// blank check the user flash area
	// the user flash area is specified as the range of addresses between
	// 'FIRST_USER_FLASH_ADDR' and 'LAST_USER_FLASH_ADDR' inclusive
	// these values are specified in 'command.h'
	
	read_datum *address;
	unsigned char blank;
	
	SendLFCR();
	SendString( "Blank checking user area..." );
	
	address = (read_datum *) FIRST_USER_FLASH_ADDR;
	
	blank = 0;
	while( address < (read_datum *) LAST_USER_FLASH_ADDR )
	{
		if ( *address != BLANK_VALUE )
		{
			blank = 1;
			// exit while loop
			break;
		}
		
		address++;
	}
	
	SendLFCR();
	
	SendString( "User area is " );
	
	if ( blank == 0 )
		SendString( "BLANK" );
	else
		SendString( "NOT blank" ); 
}

void Command_2 (void)
{
	// erase all user flash blocks
	// the first flash block to be erased is 'FIRST_USER_FLASH_BLK' and the last
	// flash block to be erased is 'NO_OF_FLASH_BLOCKS' - 1
	// these values are defined in 'command.h' and 'flash_header.h'	
	
	unsigned char uc;
	unsigned char Status;
	union union_c2s c2s;
	unsigned long *ptr;				
	pt2FunctionErase fp;			// function pointer for calling RAM based function
	
	SendLFCR();
	
	SendString( "Really erase ALL user blocks (Y/N)?" );
	c2s.us = GetByte( 10000 );
	SendByte( c2s.uc[1] );
	if ( c2s.uc[0] == TIMEOUT )
	{
		SendLFCR();
		SendString( "Timed out waiting for response" );
		return;
	}
	else if ( c2s.uc[0] == ERROR )
	{
		SendLFCR();
		SendString( "Comms error" );
		return;
	}
	else if ( !( ( c2s.uc[1] == 'y' ) || ( c2s.uc[1] == 'Y' ) ) )
	{
		SendLFCR();
		SendString( "All blocks erase cancelled" );
		return;
	}

	for( uc=FIRST_USER_FLASH_BLK; uc<NO_OF_FLASH_BLOCKS; uc++ )
	{
		SendLFCR();
		SendString( "Erasing block " );
		if( uc < 10 )
		{
			SendByte( uc + 48 );
		}
		else
		{
			SendByte( ( uc / 10 ) + 48 );
			SendByte( ( uc % 10 ) + 48 );
		}
		
		SendString ( "..." );
		// load pointer to erase block function from function pointer table
		ptr = (unsigned long *) ( INT_RAM_BASE_ADDRESS + ( ERASEBLOCK_OFFSET * 4 ) );
		fp = (pt2FunctionErase) *ptr;
		
		// call the erase function with the block to be erased and a dummy value (0)
		Status = fp ( uc, 0 );
		if ( Status == ERASE_FAIL )
		{
			SendString( "Erasing of block " );
			SendByte( uc + 48 );
			SendString( " FAILED" );
			SendLFCR();
			return;
		}
	}

	SendLFCR();
	SendString( "All user blocks erased" );
}

void Command_3 (void)
{
	// erase the specified flash block
	union union_c2s c2s;
	unsigned char BlockNum, Count;
	unsigned long *ptr;
	pt2FunctionErase fp;					// function pointer for calling RAM based function
	
	BlockNum = 0;
	
	SendLFCR();
	SendString( "Enter block no to be erased as 2-digit number e.g. 08" );
	SendLFCR();
	SendString( ">D'" );
	
	for (Count=0; Count<2; Count++)
	{
		c2s.us = GetByte( 10000 );
		SendByte( c2s.uc[1] );
		if ( c2s.uc[0] == TIMEOUT )
		{
			SendLFCR();
			SendString( "Timed out waiting for block number..." );
			return;
		}
		else if ( !isdigit( c2s.uc[1] ) )
		{
			SendLFCR();
			SendByte( CARRIAGE_RETURN );
			SendString( "Invalid character" );
			return;
		}
		else
		{
			if ( Count == 0 )
			{
				BlockNum = 10 * ( c2s.uc[1] - 48 );
			}
			else
			{
				BlockNum += ( c2s.uc[1] - 48 );
			}
		}
	}
	
	if ( !( ( BlockNum >= FIRST_USER_FLASH_BLK ) && ( BlockNum < NO_OF_FLASH_BLOCKS ) ) )
	{
		SendLFCR();
		SendString( "Invalid block number" );
		return;
	}
	
	SendLFCR();
	SendString( "Erase block (Y/N)?" );
	c2s.us = GetByte( 10000 );
	SendByte( c2s.uc[1] );
	if ( c2s.uc[0] == TIMEOUT )
	{
		SendLFCR();
		SendString( "Timed out waiting for confirmation" );
		return;
	}
	else if (  ( c2s.uc[1] == 'y' ) || ( c2s.uc[1] == 'Y' ) )
	{
		// load pointer with address of the erase function from the function pointer table
		ptr = (unsigned long *) ( INT_RAM_BASE_ADDRESS + ( ERASEBLOCK_OFFSET * 4 ) );
		fp = (pt2FunctionErase) *ptr;
		
		// call the erase function with the block to be erased and a dummy value (0)
		Count = fp ( BlockNum, 0 );
		SendLFCR();
		SendString( "Erasing of block " );
		SendByte( BlockNum + 48 );
		if ( Count == ERASE_FAIL )
		{
			SendString( " FAILED" );
		} 
		else
		{
			SendString( " PASSED" );
		}
		return;
	}
	else
	{
		SendLFCR();
		SendString( "Block erase cancelled" );
		return;
	}
}

void Command_4 (void)
{
    // download data using xmodem and program into flash
    unsigned char Status, Shift;
	volatile unsigned long Address, ul;
	union union_c2s c2s;
	
	// suggest download address as the first user flash address as usually this will be the case
	SendLFCR();
	SendString( "Download to address 0x" );
	SendString( FIRST_USER_FL_ADDR_TXT );
	SendString( " (Y/N)?" );
	c2s.us = GetByte( 10000 );
	SendByte( c2s.uc[1] );
	SendLFCR();
	if ( c2s.uc[0] == TIMEOUT )
	{
		SendLFCR();
		SendString( "Timed out waiting for confirmation..." );
		return;
	}
	else if (  ( c2s.uc[1] == 'y' ) || ( c2s.uc[1] == 'Y' ) )
	{
		Address = FIRST_USER_FLASH_ADDR;
	}
	else 
	{
		SendString( "Enter Flash memory base address as a 32-bit hex value e.g. AABBCCDD" );
	
		Address = 0;
		Shift = 28;
		
		SendLFCR();
		SendString( ">H'" );
		for (Status=0; Status<8; Status++)
		{		
			c2s.us = GetByte( 10000 );
			if ( c2s.uc[0] == TIMEOUT )
			{
				SendLFCR();
				SendString( "Timed out waiting for address..." );
				return;
			}
			else if ( c2s.uc[0] == ERROR )
			{
				SendLFCR();
				SendString( "Comms error" );
				return;
			}
			else
			{
				if ( !isxdigit( c2s.uc[1] ) )
				{
					SendLFCR();
					SendString( "Invalid character" );
					return;
				}
				SendByte( c2s.uc[1] );
				c2s.uc[1] = toupper( c2s.uc[1] );
				// convert ASCII number to hexadecimal value
				c2s.uc[1] =  isdigit( c2s.uc[1] ) ? c2s.uc[1] - 48 : c2s.uc[1] - 55;
				ul = (unsigned long) c2s.uc[1];
				ul = ul << Shift;
				Address |= ul;
				Shift -= 4;
			}		
		}
		
		SendLFCR();
	}
	
	if ( ( Address < FIRST_USER_FLASH_ADDR ) || ( Address > ( LAST_USER_FLASH_ADDR - FLASH_LINE_SIZE - 1 ) ) )
	{
		SendString( "Invalid start address - aborting" );
		return;
	}
	
	if ( ( Address % FLASH_LINE_SIZE ) != 0 )
	{
		SendString( "Start address not on a flash line boundary - aborting" );
		return;
	}
		
	SendString( "Program Flash (Y/N)?" );
	c2s.us = GetByte( 10000 );
	SendByte( c2s.uc[1] );
	if ( c2s.uc[0] == TIMEOUT )
	{
		SendLFCR();
		SendString( "Timed out waiting for response" );
		return;
	}
	else if ( c2s.uc[0] == ERROR )
	{
		SendLFCR();
		SendString( "Comms error" );
		return;
	}
	else if ( !( ( c2s.uc[1] == 'y' ) || ( c2s.uc[1] == 'Y' ) ) )
	{
		SendLFCR();
		SendString( "Flash programming cancelled" );
		return;
	}
	
	SendLFCR();
	SendString( "Start XModem download..." );
	SendLFCR();
	
	// start xmodem download and program the flash with the downloaded data
	Status = XmodemDownloadAndProgramFlash( Address );
		
	if ( Status == XM_TIMEOUT )
		SendString( "Timeout" );
	else if ( Status == XM_OK )
		SendString( "Download OK" );
	else if ( Status == XM_PROG_FAIL )
		SendString( "Flash program FAIL" );

	SendLFCR();
	
	PurgeComms( 1000 );
}

void Command_5 (void)
{
	// call the target application
	union union_c2s c2s;
	unsigned long *ptr;
	pt2FunctionErase fp;
	
	SendLFCR();
	SendString( "Run user program (Y/N)?" );
	c2s.us = GetByte( 10000 );
	SendByte( c2s.uc[1] );
	if ( c2s.uc[0] == TIMEOUT )
	{
		SendLFCR();
		SendString( "Timed out waiting for response" );
		return;
	}
	else if ( c2s.uc[0] == ERROR )
	{
		SendLFCR();
		SendString( "Comms error" );
		return;
	}
	else if ( !( ( c2s.uc[1] == 'y' ) || ( c2s.uc[1] == 'Y' ) ) )
	{
		SendLFCR();
		SendString( "Run user program request cancelled" );
		return;
	}
	
	SendLFCR();
	
	// run user program
	// check user reset address is not all erased
	ptr = (unsigned long *) ( FIRST_USER_FLASH_ADDR + RESET_VECTOR_OFFSET );
	if ( *ptr != 0xffffffff )
	{
		SendString( "Running user program..." );
		// call user program and never return
		fp = (pt2FunctionErase) *ptr;
		fp( 0, 0);
	}
	else
	{
		// invalid user reset address
		SendString( "Invalid user reset address.  Staying put." );
		return;
	}
}
