// flash.c

#include <intrinsics.h>
#include "flash_header.h"
#include "..\xmodemboot\iom16c62p.h"

//
//	__root directive required to ensure that the function and variable/constant
//	data is included in the object code even if the linker thinks it is not used
//

typedef unsigned char (*pt2Function)( unsigned long, unsigned short * );

#pragma location="CPROGERA"
__root const pt2Function fp[] = {
	BlockErase,
	Program_128_Bytes
};

__root const unsigned long Block_Address[][2] = {
	{ 0xFF000, 0xFFFFE },
	{ 0xFE000, 0xFEFFE },
	{ 0xFC000, 0xFDFFE },
	{ 0xFA000, 0xFBFFE },
	{ 0xF8000, 0xF9FFE },
	{ 0xF0000, 0xF7FFE },
	{ 0xE0000, 0xEFFFE },
	{ 0xD0000, 0xDFFFE },
	{ 0xC0000, 0xCFFFE },
	{ 0xB0000, 0xBFFFE },
	{ 0xA0000, 0xAFFFE },
	{ 0x0F000, 0x0FFFE }
};

__root void EnterEW0Mode( void )
{
	// during flash memory programming, erasing etc
	// the CPU clock frequency should be < 10MHz (with 1 wait state)
	// or < 6.5MHz (no wait state)
	// 3D board has a 6MHz xtal which is internal multiplied by 4 and divided by 0
	// therefore this clock should be divided by 4 during flash memory operations
	
	__disable_interrupt();
	
	// enable access to the clock mode register 1 (CM0)
	PRCR_bit.PRC0 = 1;
	
	// CM1 - System Clock Control Register 1
	//-------------------------------------------------------------------------
  	CM1_bit.CM17 = 1;		// Main Clock Divide Ratio
	CM1_bit.CM16 = 0;	  	//	CM17	CM16
							//	0		0		- No Division
							//	0		1		- Divide by 2
							//	1		0		- Divide by 4
							//	1		1		- Divide by 16
	
	// disable access to the clock mode register 1 (CM0)
	PRCR_bit.PRC0 = 0;
	
	// CPU now operating at 6MHz
	//
	// enter CPU re-write mode
	FMR0_bit.FMR01 = 0;		// first write 0 then write 1
	FMR0_bit.FMR01 = 1;
	
	// set EW0 mode
	FMR1_bit.FMR11 = 0;
	
	// disable all lock bit protection to allow programming and erase operations
	FMR0_bit.FMR02 = 0;		// first write 0 then write 1
	FMR0_bit.FMR02 = 1;
	
	// all flash blocks now unlocked
	//
	// wait while flash is busy
	while( FMR0_bit.FMR00 == 0 );
}

__root void ExitEW0Mode( void )
{
	// when this function exits the CPU clock will be 24MHz
	
	// enable the lock-bit protection
	FMR0_bit.FMR02 = 0;
	
	// exit CPU re-write mode
	FMR0_bit.FMR01 = 0;
	
	// enable access to the clock mode register 1 (CM0)
	PRCR_bit.PRC0 = 1;
	
	// CM1 - System Clock Control Register 1
	//-------------------------------------------------------------------------
  	CM1_bit.CM17 = 0;		// Main Clock Divide Ratio
	CM1_bit.CM16 = 0;	  	//	CM17	CM16
							//	0		0		- No Division
							//	0		1		- Divide by 2
							//	1		0		- Divide by 4
							//	1		1		- Divide by 16
	
	// disable access to the clock mode register 1 (CM0)
	PRCR_bit.PRC0 = 0;
	
	__enable_interrupt();
}

__root unsigned char BlockErase( unsigned long BlockNum, unsigned short *Dummy )
{
	// attempts to erase the block specified by BlockNum
	//
	// returns:
	//
	// ERASE_PASS
	// or
	// ERASE_FAIL
	
	volatile unsigned short __data20 *p;
	volatile unsigned char Status;
	unsigned long ul;
	volatile unsigned short Tmp;
	
	// check if block needs erasing
	Status = 0;
	p = (volatile unsigned short __data20 *)Block_Address[ BlockNum ][ 0 ];
	for( ul = Block_Address[ BlockNum ][ 0 ]; ul<Block_Address[ BlockNum ][ 1 ] + 1; ul++ )
	{
		if( *p != 0xffff )
		{
			Status = 1;
			break;
		}
		else
		{
			p++;
		}
	}
	
	if( Status == 0 ) return ERASE_PASS;
	
	EnterEW0Mode();
		
	// write erase block command to even flash address
	p = (volatile unsigned short __data20 *)Block_Address[ BlockNum ][ 0 ];
	*p = BLOCK_ERASE;
	// write H'xxD0 to upper most address in block to be erased
	p = (volatile unsigned short __data20 *)Block_Address[ BlockNum ][ 1 ];
	*p = 0x00D0;
	
	// wait while flash is busy
	while( FMR0_bit.FMR00 == 0 );
	
	// read flash to allow access to the status register
	Tmp = *p;
	
	// read the erase status flag in the status register
	Status = FMR0_bit.FMR07;
	
	ExitEW0Mode();
	
	if( Status == 0 )
	{
		// flash block erase successful
		return ERASE_PASS;
	}
	else
	{
		// flash block erase failed
		return ERASE_FAIL;
	}
}

__root unsigned char Program_128_Bytes( unsigned long Address, unsigned short *Data )
{
	// attempts to program 128 bytes (in word units)
	// at address the specified address
	//
	// returns:
	// 
	// PROG_PASS
	// or
	// PROG_FAIL
	
	unsigned char Count, Status;
	unsigned long Addr;
	volatile unsigned short __data20 *p;
	volatile unsigned short Tmp;
	
	Count = 0;
	Addr = Address;
	
	EnterEW0Mode();
	
	Status = PROG_PASS;
	while( ( Count < (128 / 2) ) && ( Status == PROG_PASS ) )
	{
		p = (volatile unsigned short __data20 *)Addr;
		*p = PROGRAM_CMD;
		*p = Data[ Count ];
		
		// wait while flash is busy
		while( FMR0_bit.FMR00 == 0 );
	
		// read flash to allow access to the status register
		Tmp = *p;
		
		// read the erase status flag in the status register
		Status = FMR0_bit.FMR06;
				
		if( Status == 0 )
		{
			Status = PROG_PASS;
		}
		else
		{
			Status = PROG_FAIL;
		}
		
		Count++;
		Addr += 2;
	}

	ExitEW0Mode();
	
	return Status;
}
